@@ -9,14 +9,9 @@
 
 #include "config.h"
 
-#ifndef lint
-static const char sccsid[] = "$Id: ex.c,v 10.75 2004/03/16 14:13:35 skimo Exp $ (Berkeley) $Date: 2004/03/16 14:13:35 $";
-#endif /* not lint */
-
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 
 #include <bitstring.h>
 #include <ctype.h>
@@ -32,34 +27,32 @@ static const char sccsid[] = "$Id: ex.c,v 10.75 2004/03/16 14:13:35 skimo Exp $
 #include "../vi/vi.h"
 
 #if defined(DEBUG) && defined(COMLOG)
-static void	ex_comlog __P((SCR *, EXCMD *));
+static void	ex_comlog(SCR *, EXCMD *);
 #endif
 static EXCMDLIST const *
-		ex_comm_search __P((SCR *, CHAR_T *, size_t));
-static int	ex_discard __P((SCR *));
-static int	ex_line __P((SCR *, EXCMD *, MARK *, int *, int *));
-static int	ex_load __P((SCR *));
-static void	ex_unknown __P((SCR *, CHAR_T *, size_t));
+		ex_comm_search(CHAR_T *, size_t);
+static int	ex_discard(SCR *);
+static int	ex_line(SCR *, EXCMD *, MARK *, int *, int *);
+static int	ex_load(SCR *);
+static void	ex_unknown(SCR *, CHAR_T *, size_t);
 
 /*
  * ex --
  *	Main ex loop.
  *
- * PUBLIC: int ex __P((SCR **));
+ * PUBLIC: int ex(SCR **);
  */
 int
 ex(SCR **spp)
 {
 	EX_PRIVATE *exp;
 	GS *gp;
-	WIN *wp;
 	MSGS *mp;
 	SCR *sp;
 	TEXT *tp;
 	u_int32_t flags;
 
 	sp = *spp;
-	wp = sp->wp;
 	gp = sp->gp;
 	exp = EXP(sp);
 
@@ -68,17 +61,17 @@ ex(SCR **spp)
 		return (1);
 
 	/* Flush any saved messages. */
-	while ((mp = gp->msgq.lh_first) != NULL) {
-		wp->scr_msg(sp, mp->mtype, mp->buf, mp->len);
-		LIST_REMOVE(mp, q);
+	while ((mp = SLIST_FIRST(gp->msgq)) != NULL) {
+		gp->scr_msg(sp, mp->mtype, mp->buf, mp->len);
+		SLIST_REMOVE_HEAD(gp->msgq, q);
 		free(mp->buf);
 		free(mp);
 	}
 
 	/* If reading from a file, errors should have name and line info. */
 	if (F_ISSET(gp, G_SCRIPTED)) {
-		wp->excmd.if_lno = 1;
-		wp->excmd.if_name = "script";
+		gp->excmd.if_lno = 1;
+		gp->excmd.if_name = "script";
 	}
 
 	/*
@@ -89,7 +82,7 @@ ex(SCR **spp)
 	 * "^H discarded", that was displayed.  We don't bother.
 	 */
 	LF_INIT(TXT_BACKSLASH | TXT_CNTRLD | TXT_CR);
-	for (;; ++wp->excmd.if_lno) {
+	for (;; ++gp->excmd.if_lno) {
 		/* Display status line and flush. */
 		if (F_ISSET(sp, SC_STATUS)) {
 			if (!F_ISSET(sp, SC_EX_SILENT))
@@ -106,7 +99,7 @@ ex(SCR **spp)
 
 		/* Clear any current interrupts, and get a command. */
 		CLR_INTERRUPT(sp);
-		if (ex_txt(sp, &sp->tiq, ':', flags))
+		if (ex_txt(sp, sp->tiq, ':', flags))
 			return (1);
 		if (INTERRUPTED(sp)) {
 			(void)ex_puts(sp, "\n");
@@ -115,22 +108,21 @@ ex(SCR **spp)
 		}
 
 		/* Initialize the command structure. */
-		CLEAR_EX_PARSER(&wp->excmd);
+		CLEAR_EX_PARSER(&gp->excmd);
 
 		/*
 		 * If the user entered a single carriage return, send
 		 * ex_cmd() a separator -- it discards single newlines.
 		 */
-		tp = sp->tiq.cqh_first;
+		tp = TAILQ_FIRST(sp->tiq);
 		if (tp->len == 0) {
-			static CHAR_T space = ' ';
-			wp->excmd.cp = &space;	/* __TK__ why not |? */
-			wp->excmd.clen = 1;
+			gp->excmd.cp = L(" ");	/* __TK__ why not |? */
+			gp->excmd.clen = 1;
 		} else {
-			wp->excmd.cp = tp->lb;
-			wp->excmd.clen = tp->len;
+			gp->excmd.cp = tp->lb;
+			gp->excmd.clen = tp->len;
 		}
-		F_INIT(&wp->excmd, E_NRSEP);
+		F_INIT(&gp->excmd, E_NRSEP);
 
 		if (ex_cmd(sp) && F_ISSET(gp, G_SCRIPTED))
 			return (1);
@@ -162,6 +154,10 @@ ex(SCR **spp)
 			if (file_end(sp, NULL, F_ISSET(sp, SC_EXIT_FORCE)))
 				return (1);
 			*spp = screen_next(sp);
+			if (*spp) {
+				F_CLR(*spp, SC_SCR_VI);
+				F_SET(*spp, SC_SCR_EX);
+			}
 			return (screen_end(sp));
 		}
 	}
@@ -191,7 +187,7 @@ ex(SCR **spp)
  *
  * For extra credit, try them in a startup .exrc file.
  *
- * PUBLIC: int ex_cmd __P((SCR *));
+ * PUBLIC: int ex_cmd(SCR *);
  */
 int
 ex_cmd(SCR *sp)
@@ -200,9 +196,8 @@ ex_cmd(SCR *sp)
 	EX_PRIVATE *exp;
 	EXCMD *ecp;
 	GS *gp;
-	WIN *wp;
 	MARK cur;
-	db_recno_t lno;
+	recno_t lno;
 	size_t arg1_len, discard, len;
 	u_int32_t flags;
 	long ltmp;
@@ -210,12 +205,11 @@ ex_cmd(SCR *sp)
 	int cnt, delim, isaddr, namelen;
 	int newscreen, notempty, tmp, vi_address;
 	CHAR_T *arg1, *s, *p, *t;
-	CHAR_T ch;
+	CHAR_T ch = '\0';
 	CHAR_T *n;
 	char *np;
 
 	gp = sp->gp;
-	wp = sp->wp;
 	exp = EXP(sp);
 
 	/*
@@ -223,12 +217,12 @@ ex_cmd(SCR *sp)
 	 * This means that *everything* must be resolved when we leave
 	 * this function for any reason.
 	 */
-loop:	ecp = wp->ecq.lh_first;
+loop:	ecp = SLIST_FIRST(gp->ecq);
 
 	/* If we're reading a command from a file, set up error information. */
 	if (ecp->if_name != NULL) {
-		wp->if_lno = ecp->if_lno;
-		wp->if_name = ecp->if_name;
+		gp->if_lno = ecp->if_lno;
+		gp->if_name = ecp->if_name;
 	}
 
 	/*
@@ -244,7 +238,7 @@ loop:	ecp = wp->ecq.lh_first;
 
 	/* If we found a newline, increment the count now. */
 	if (F_ISSET(ecp, E_NEWLINE)) {
-		++wp->if_lno;
+		++gp->if_lno;
 		++ecp->if_lno;
 		F_CLR(ecp, E_NEWLINE);
 	}
@@ -263,9 +257,9 @@ loop:	ecp = wp->ecq.lh_first;
 	/* Skip <blank>s, empty lines.  */
 	for (notempty = 0; ecp->clen > 0; ++ecp->cp, --ecp->clen)
 		if ((ch = *ecp->cp) == '\n') {
-			++wp->if_lno;
+			++gp->if_lno;
 			++ecp->if_lno;
-		} else if (ISBLANK(ch))
+		} else if (cmdskip(ch))
 			notempty = 1;
 		else
 			break;
@@ -304,7 +298,7 @@ loop:	ecp = wp->ecq.lh_first;
 	/* Skip whitespace. */
 	for (; ecp->clen > 0; ++ecp->cp, --ecp->clen) {
 		ch = *ecp->cp;
-		if (!ISBLANK(ch))
+		if (!cmdskip(ch))
 			break;
 	}
 
@@ -327,7 +321,7 @@ loop:	ecp = wp->ecq.lh_first;
 	    (!notempty || F_ISSET(sp, SC_VI) || F_ISSET(ecp, E_BLIGNORE))) {
 		if (ex_load(sp))
 			goto rfail;
-		ecp = wp->ecq.lh_first;
+		ecp = SLIST_FIRST(gp->ecq);
 		if (ecp->clen == 0)
 			goto rsuccess;
 		goto loop;
@@ -359,7 +353,7 @@ loop:	ecp = wp->ecq.lh_first;
 	 */
 	for (; ecp->clen > 0; ++ecp->cp, --ecp->clen) {
 		ch = *ecp->cp;
-		if (!ISBLANK(ch) && ch != ':')
+		if (!cmdskip(ch) && ch != ':')
 			break;
 	}
 
@@ -383,10 +377,10 @@ loop:	ecp = wp->ecq.lh_first;
 	 * from vi mode, and displayed lines 2, 3, and 4, so we do a default
 	 * command for each separator.
 	 */
-#define	SINGLE_CHAR_COMMANDS	"\004!#&*<=>@~"
+#define	SINGLE_CHAR_COMMANDS	L("\004!#&*<=>@~")
 	newscreen = 0;
 	if (ecp->clen != 0 && ecp->cp[0] != '|' && ecp->cp[0] != '\n') {
-		if (strchr(SINGLE_CHAR_COMMANDS, *ecp->cp)) {
+		if (STRCHR(SINGLE_CHAR_COMMANDS, *ecp->cp)) {
 			p = ecp->cp;
 			++ecp->cp;
 			--ecp->clen;
@@ -394,7 +388,7 @@ loop:	ecp = wp->ecq.lh_first;
 		} else {
 			for (p = ecp->cp;
 			    ecp->clen > 0; --ecp->clen, ++ecp->cp)
-				if (!ISALPHA(*ecp->cp))
+				if (!isazAZ(*ecp->cp))
 					break;
 			if ((namelen = ecp->cp - p) == 0) {
 				msgq(sp, M_ERR, "080|Unknown command name");
@@ -457,7 +451,7 @@ loop:	ecp = wp->ecq.lh_first;
 		 * follows the 's', but we limit the choices here to "cgr" so
 		 * that we get unknown command messages for wrong combinations.
 		 */
-		if ((ecp->cmd = ex_comm_search(sp, p, namelen)) == NULL)
+		if ((ecp->cmd = ex_comm_search(p, namelen)) == NULL)
 			switch (p[0]) {
 			case 'k':
 				if (namelen == 2) {
@@ -580,8 +574,8 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 
 	/* Check for ex mode legality. */
 	if (F_ISSET(sp, SC_EX) && (F_ISSET(ecp->cmd, E_VIONLY) || newscreen)) {
-		msgq(sp, M_ERR,
-		    "082|%s: command not available in ex mode", ecp->cmd->name);
+		msgq_wstr(sp, M_ERR, ecp->cmd->name,
+		    "082|%s: command not available in ex mode");
 		goto err;
 	}
 
@@ -651,7 +645,7 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 			ecp->save_cmd = ecp->cp;
 		}
 		for (; ecp->clen > 0; --ecp->clen, ++ecp->cp)
-			if (!ISBLANK(*ecp->cp))
+			if (!cmdskip(*ecp->cp))
 				break;
 		/*
 		 * QUOTING NOTE:
@@ -673,7 +667,7 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 					++discard;
 					--ecp->clen;
 					ch = *++ecp->cp;
-				} else if (ISBLANK(ch))
+				} else if (cmdskip(ch))
 					break;
 				*p++ = ch;
 			}
@@ -700,7 +694,7 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 				--ecp->clen;
 				ch = *++ecp->cp;
 
-				++wp->if_lno;
+				++gp->if_lno;
 				++ecp->if_lno;
 			} else if (ch == '\n')
 				break;
@@ -717,7 +711,7 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 		 */
 		for (tmp = 0; ecp->clen > 0; --ecp->clen, ++ecp->cp) {
 			ch = *ecp->cp;
-			if (ISBLANK(ch))
+			if (cmdskip(ch))
 				tmp = 1;
 			else
 				break;
@@ -735,10 +729,10 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 		 * into something like ":s g", so use the special s command.
 		 */
 		for (; ecp->clen > 0; --ecp->clen, ++ecp->cp)
-			if (!ISBLANK(ecp->cp[0]))
+			if (!cmdskip(ecp->cp[0]))
 				break;
 
-		if (ISALNUM(ecp->cp[0]) || ecp->cp[0] == '|') {
+		if (is09azAZ(ecp->cp[0]) || ecp->cp[0] == '|') {
 			ecp->rcmd = cmds[C_SUBSTITUTE];
 			ecp->rcmd.fn = ex_subagain;
 			ecp->cmd = &ecp->rcmd;
@@ -779,13 +773,14 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 	 * no longer useful.
 	 */
 	vi_address = ecp->clen != 0 && ecp->cp[0] != '\n';
+	ecp->trailing = 0;
 	for (p = ecp->cp; ecp->clen > 0; --ecp->clen, ++ecp->cp) {
 		ch = ecp->cp[0];
 		if (IS_ESCAPE(sp, ecp, ch) && ecp->clen > 1) {
 			CHAR_T tmp = ecp->cp[1];
 			if (tmp == '\n' || tmp == '|') {
 				if (tmp == '\n') {
-					++wp->if_lno;
+					++gp->if_lno;
 					++ecp->if_lno;
 				}
 				++discard;
@@ -794,6 +789,7 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 				ch = tmp;
 			}
 		} else if (ch == '\n' || ch == '|') {
+			ecp->trailing = 1;
 			if (ch == '\n')
 				F_SET(ecp, E_NEWLINE);
 			--ecp->clen;
@@ -821,10 +817,14 @@ skip_srch:	if (ecp->cmd == &cmds[C_VISUAL_EX] && F_ISSET(sp, SC_VI))
 	 * this isn't a particularly complex trap, and if backslashes were
 	 * legal in set commands, this would have to be much more complicated.
 	 */
-	if (ecp->cmd == &cmds[C_SET])
+	if (ecp->cmd == &cmds[C_SET]) {
 		for (p = ecp->cp, len = ecp->clen; len > 0; --len, ++p)
-			if (*p == '\\')
+			if (IS_ESCAPE(sp, ecp, *p) && len > 1) {
+				--len;
+				++p;
+			} else if (*p == '\\')
 				*p = CH_LITERAL;
+	}
 
 	/*
 	 * Set the default addresses.  It's an error to specify an address for
@@ -912,7 +912,7 @@ two_addr:	switch (ecp->addrcnt) {
 			ecp->addr1.cno = ecp->addr2.cno = sp->cno;
 			break;
 		case 1:				/* Default to first address. */
-			//ecp->addrcnt = 2;	/* XXX Was this needed ??? */
+			ecp->addrcnt = 2;
 			ecp->addr2 = ecp->addr1;
 			break;
 		case 2:
@@ -959,7 +959,7 @@ two_addr:	switch (ecp->addrcnt) {
 
 		/* Skip leading <blank>s. */
 		for (; ecp->clen > 0; --ecp->clen, ++ecp->cp)
-			if (!ISBLANK(*ecp->cp))
+			if (!cmdskip(*ecp->cp))
 				break;
 		if (ecp->clen == 0)
 			break;
@@ -1066,7 +1066,7 @@ end_case23:		break;
 			else if (*ecp->cp == '+')
 				FL_SET(ecp->iflags, E_C_COUNT_POS);
 			if ((nret =
-			    nget_slong(sp, &ltmp, ecp->cp, &t, 10)) != NUM_OK) {
+			    nget_slong(&ltmp, ecp->cp, &t, 10)) != NUM_OK) {
 				ex_badaddr(sp, NULL, A_NOTSET, nret);
 				goto err;
 			}
@@ -1156,7 +1156,7 @@ end_case23:		break;
 				    ecp, ch) && ecp->clen > 1) {
 					--ecp->clen;
 					*p++ = *++ecp->cp;
-				} else if (ISBLANK(ch)) {
+				} else if (cmdskip(ch)) {
 					++ecp->cp;
 					--ecp->clen;
 					break;
@@ -1170,7 +1170,7 @@ end_case23:		break;
 			for (; ecp->clen > 0;
 			    --ecp->clen, ++ecp->cp) {
 				ch = *ecp->cp;
-				if (!ISBLANK(ch))
+				if (!cmdskip(ch))
 					break;
 			}
 			if (ecp->clen == 0)
@@ -1205,17 +1205,23 @@ arg_cnt_chk:		if (*++np != 'N') {		/* N */
 					goto usage;
 			}
 			goto addr_verify;
-		default:
+		default: {
+			size_t nlen;
+			char *nstr;
+
+			INT2CHAR(sp, ecp->cmd->name, STRLEN(ecp->cmd->name) + 1,
+			    nstr, nlen);
 			msgq(sp, M_ERR,
 			    "085|Internal syntax table error (%s: %s)",
-			    ecp->cmd->name, KEY_NAME(sp, *np));
+			    nstr, KEY_NAME(sp, *np));
+		}
 		}
 	}
 
 	/* Skip trailing whitespace. */
 	for (; ecp->clen > 0; --ecp->clen) {
 		ch = *ecp->cp++;
-		if (!ISBLANK(ch))
+		if (!cmdskip(ch))
 			break;
 	}
 
@@ -1385,10 +1391,10 @@ addr_verify:
 
 #ifdef DEBUG
 	/* Make sure no function left global temporary space locked. */
-	if (F_ISSET(wp, W_TMP_INUSE)) {
-		F_CLR(wp, W_TMP_INUSE);
-		msgq(sp, M_ERR, "087|%s: temporary buffer not released",
-		    ecp->cmd->name);
+	if (F_ISSET(gp, G_TMP_INUSE)) {
+		F_CLR(gp, G_TMP_INUSE);
+		msgq_wstr(sp, M_ERR, ecp->cmd->name,
+		    "087|%s: temporary buffer not released");
 	}
 #endif
 	/*
@@ -1424,7 +1430,7 @@ addr_verify:
 				goto err;
 			}
 		} else {
-			if (!NPFITS(DB_MAX_RECORDS, sp->lno, ecp->flagoff)) {
+			if (!NPFITS(MAX_REC_NUMBER, sp->lno, ecp->flagoff)) {
 				ex_badaddr(sp, NULL, A_NOTSET, NUM_OVER);
 				goto err;
 			}
@@ -1463,8 +1469,13 @@ addr_verify:
 		LF_INIT(FL_ISSET(ecp->iflags, E_C_HASH | E_C_LIST | E_C_PRINT));
 		if (!LF_ISSET(E_C_HASH | E_C_LIST | E_C_PRINT | E_NOAUTO) &&
 		    !F_ISSET(sp, SC_EX_GLOBAL) &&
-		    O_ISSET(sp, O_AUTOPRINT) && F_ISSET(ecp, E_AUTOPRINT))
-			LF_INIT(E_C_PRINT);
+		    O_ISSET(sp, O_AUTOPRINT) && F_ISSET(ecp, E_AUTOPRINT)) {
+			/* Honor the number option if autoprint is set. */
+			if (F_ISSET(ecp, E_OPTNUM))
+				LF_INIT(E_C_HASH);
+			else
+				LF_INIT(E_C_PRINT);
+		}
 
 		if (LF_ISSET(E_C_HASH | E_C_LIST | E_C_PRINT)) {
 			cur.lno = sp->lno;
@@ -1501,7 +1512,7 @@ addr_verify:
 
 		ecp->save_cmd -= arg1_len;
 		ecp->save_cmdlen += arg1_len;
-		MEMCPYW(ecp->save_cmd, arg1, arg1_len);
+		MEMMOVE(ecp->save_cmd, arg1, arg1_len);
 
 		/*
 		 * Any commands executed from a +cmd are executed starting at
@@ -1535,8 +1546,7 @@ addr_verify:
 	 */
 	if (F_ISSET(sp, SC_EXIT | SC_EXIT_FORCE | SC_FSWITCH | SC_SSWITCH)) {
 		at_found = gv_found = 0;
-		for (ecp = wp->ecq.lh_first;
-		    ecp != NULL; ecp = ecp->q.le_next)
+		SLIST_FOREACH(ecp, sp->gp->ecq, q)
 			switch (ecp->agv_flags) {
 			case 0:
 			case AGV_AT_NORANGE:
@@ -1588,7 +1598,7 @@ err:	/*
 				break;
 			}
 		}
-	if (ecp->save_cmdlen != 0 || wp->ecq.lh_first != &wp->excmd) {
+	if (ecp->save_cmdlen != 0 || SLIST_FIRST(gp->ecq) != &gp->excmd) {
 discard:	msgq(sp, M_BERR,
 		    "092|Ex command failed: pending commands discarded");
 		ex_discard(sp);
@@ -1602,7 +1612,7 @@ rfail:	tmp = 1;
 rsuccess:	tmp = 0;
 
 	/* Turn off any file name error information. */
-	wp->if_name = NULL;
+	gp->if_name = NULL;
 
 	/* Turn off the global bit. */
 	F_CLR(sp, SC_EX_GLOBAL);
@@ -1614,7 +1624,7 @@ rsuccess:	tmp = 0;
  * ex_range --
  *	Get a line range for ex commands, or perform a vi ex address search.
  *
- * PUBLIC: int ex_range __P((SCR *, EXCMD *, int *));
+ * PUBLIC: int ex_range(SCR *, EXCMD *, int *);
  */
 int
 ex_range(SCR *sp, EXCMD *ecp, int *errp)
@@ -1687,12 +1697,12 @@ ex_range(SCR *sp, EXCMD *ecp, int *errp)
 			++ecp->cp;
 			--ecp->clen;
 			break;
-		case ',':               /* Comma delimiter. */
+		case ',':	       /* Comma delimiter. */
 			/* Vi ex address searches didn't permit commas. */
 			if (F_ISSET(ecp, E_VISEARCH))
 				goto ret;
 			/* FALLTHROUGH */
-		case ';':               /* Semi-colon delimiter. */
+		case ';':	       /* Semi-colon delimiter. */
 			if (sp->ep == NULL) {
 				ex_badaddr(sp, NULL, A_EMPTY, NUM_OK);
 				*errp = 1;
@@ -1823,7 +1833,7 @@ ex_line(SCR *sp, EXCMD *ecp, MARK *mp, int *isaddrp, int *errp)
 	GS *gp;
 	long total, val;
 	int isneg;
-	int (*sf) __P((SCR *, MARK *, MARK *, CHAR_T *, size_t, CHAR_T **, u_int));
+	int (*sf)(SCR *, MARK *, MARK *, CHAR_T *, size_t, CHAR_T **, u_int);
 	CHAR_T *endp;
 
 	gp = sp->gp;
@@ -1833,7 +1843,7 @@ ex_line(SCR *sp, EXCMD *ecp, MARK *mp, int *isaddrp, int *errp)
 	F_CLR(ecp, E_DELTA);
 
 	/* No addresses permitted until a file has been read in. */
-	if (sp->ep == NULL && strchr("$0123456789'\\/?.+-^", *ecp->cp)) {
+	if (sp->ep == NULL && STRCHR(L("$0123456789'\\/?.+-^"), *ecp->cp)) {
 		ex_badaddr(sp, NULL, A_EMPTY, NUM_OK);
 		*errp = 1;
 		return (0);
@@ -1855,12 +1865,12 @@ ex_line(SCR *sp, EXCMD *ecp, MARK *mp, int *isaddrp, int *errp)
 		*isaddrp = 1;
 		F_SET(ecp, E_ABSMARK);
 
-		if ((nret = nget_slong(sp, &val, ecp->cp, &endp, 10)) != NUM_OK) {
+		if ((nret = nget_slong(&val, ecp->cp, &endp, 10)) != NUM_OK) {
 			ex_badaddr(sp, NULL, A_NOTSET, nret);
 			*errp = 1;
 			return (0);
 		}
-		if (!NPFITS(DB_MAX_RECORDS, 0, val)) {
+		if (!NPFITS(MAX_REC_NUMBER, 0, val)) {
 			ex_badaddr(sp, NULL, A_NOTSET, NUM_OVER);
 			*errp = 1;
 			return (0);
@@ -1894,7 +1904,7 @@ ex_line(SCR *sp, EXCMD *ecp, MARK *mp, int *isaddrp, int *errp)
 		 * difference.  C'est la vie.
 		 */
 		if (ecp->clen < 2 ||
-		    ecp->cp[1] != '/' && ecp->cp[1] != '?') {
+		    (ecp->cp[1] != '/' && ecp->cp[1] != '?')) {
 			msgq(sp, M_ERR, "096|\\ not followed by / or ?");
 			*errp = 1;
 			return (0);
@@ -1955,7 +1965,7 @@ search:		mp->lno = sp->lno;
 
 	/* Skip trailing <blank>s. */
 	for (; ecp->clen > 0 &&
-	    ISBLANK(ecp->cp[0]); ++ecp->cp, --ecp->clen);
+	    cmdskip(ecp->cp[0]); ++ecp->cp, --ecp->clen);
 
 	/*
 	 * Evaluate any offset.  If no address yet found, the offset
@@ -1998,11 +2008,11 @@ search:		mp->lno = sp->lno;
 		 */
 		F_SET(ecp, E_DELTA);
 		for (;;) {
-			for (; ecp->clen > 0 && ISBLANK(ecp->cp[0]);
+			for (; ecp->clen > 0 && cmdskip(ecp->cp[0]);
 			    ++ecp->cp, --ecp->clen);
-			if (ecp->clen == 0 || !ISDIGIT(ecp->cp[0]) &&
+			if (ecp->clen == 0 || (!ISDIGIT(ecp->cp[0]) &&
 			    ecp->cp[0] != '+' && ecp->cp[0] != '-' &&
-			    ecp->cp[0] != '^')
+			    ecp->cp[0] != '^'))
 				break;
 			if (!ISDIGIT(ecp->cp[0]) &&
 			    !ISDIGIT(ecp->cp[1])) {
@@ -2019,7 +2029,7 @@ search:		mp->lno = sp->lno;
 					isneg = 0;
 
 				/* Get a signed long, add it to the total. */
-				if ((nret = nget_slong(sp, &val,
+				if ((nret = nget_slong(&val,
 				    ecp->cp, &endp, 10)) != NUM_OK ||
 				    (nret = NADD_SLONG(sp,
 				    total, val)) != NUM_OK) {
@@ -2036,7 +2046,7 @@ search:		mp->lno = sp->lno;
 
 	/*
 	 * Any value less than 0 is an error.  Make sure that the new value
-	 * will fit into a db_recno_t.
+	 * will fit into a recno_t.
 	 */
 	if (*isaddrp && total != 0) {
 		if (total < 0) {
@@ -2047,7 +2057,7 @@ search:		mp->lno = sp->lno;
 				return (0);
 			}
 		} else
-			if (!NPFITS(DB_MAX_RECORDS, mp->lno, total)) {
+			if (!NPFITS(MAX_REC_NUMBER, mp->lno, total)) {
 				ex_badaddr(sp, NULL, A_NOTSET, NUM_OVER);
 				*errp = 1;
 				return (0);
@@ -2065,7 +2075,7 @@ search:		mp->lno = sp->lno;
 static int
 ex_load(SCR *sp)
 {
-	WIN *wp;
+	GS *gp;
 	EXCMD *ecp;
 	RANGE *rp;
 
@@ -2075,17 +2085,19 @@ ex_load(SCR *sp)
 	 * Lose any exhausted commands.  We know that the first command
 	 * can't be an AGV command, which makes things a bit easier.
 	 */
-	for (wp = sp->wp;;) {
+	for (gp = sp->gp;;) {
+		ecp = SLIST_FIRST(gp->ecq);
+
+		/* Discard the allocated source name as requested. */
+		if (F_ISSET(ecp, E_NAMEDISCARD))
+			free(ecp->if_name);
+
 		/*
 		 * If we're back to the original structure, leave it around,
-		 * but discard any allocated source name, we've returned to
-		 * the beginning of the command stack.
+		 * since we've returned to the beginning of the command stack.
 		 */
-		if ((ecp = wp->ecq.lh_first) == &wp->excmd) {
-			if (F_ISSET(ecp, E_NAMEDISCARD)) {
-				free(ecp->if_name);
-				ecp->if_name = NULL;
-			}
+		if (ecp == &gp->excmd) {
+			ecp->if_name = NULL;
 			return (0);
 		}
 
@@ -2105,20 +2117,20 @@ ex_load(SCR *sp)
 		 */
 		if (FL_ISSET(ecp->agv_flags, AGV_ALL)) {
 			/* Discard any exhausted ranges. */
-			while ((rp = ecp->rq.cqh_first) != (void *)&ecp->rq)
+			while ((rp = TAILQ_FIRST(ecp->rq)) != NULL)
 				if (rp->start > rp->stop) {
-					CIRCLEQ_REMOVE(&ecp->rq, rp, q);
+					TAILQ_REMOVE(ecp->rq, rp, q);
 					free(rp);
 				} else
 					break;
 
 			/* If there's another range, continue with it. */
-			if (rp != (void *)&ecp->rq)
+			if (rp != NULL)
 				break;
 
 			/* If it's a global/v command, fix up the last line. */
 			if (FL_ISSET(ecp->agv_flags,
-			    AGV_GLOBAL | AGV_V) && ecp->range_lno != OOBLNO)
+			    AGV_GLOBAL | AGV_V) && ecp->range_lno != OOBLNO) {
 				if (db_exist(sp, ecp->range_lno))
 					sp->lno = ecp->range_lno;
 				else {
@@ -2127,11 +2139,12 @@ ex_load(SCR *sp)
 					if (sp->lno == 0)
 						sp->lno = 1;
 				}
+			}
 			free(ecp->o_cp);
 		}
 
 		/* Discard the EXCMD. */
-		LIST_REMOVE(ecp, q);
+		SLIST_REMOVE_HEAD(gp->ecq, q);
 		free(ecp);
 	}
 
@@ -2142,7 +2155,7 @@ ex_load(SCR *sp)
 	 * so we have play games.
 	 */
 	ecp->cp = ecp->o_cp;
-	MEMCPYW(ecp->cp, ecp->cp + ecp->o_clen, ecp->o_clen);
+	MEMCPY(ecp->cp, ecp->cp + ecp->o_clen, ecp->o_clen);
 	ecp->clen = ecp->o_clen;
 	ecp->range_lno = sp->lno = rp->start++;
 
@@ -2158,7 +2171,7 @@ ex_load(SCR *sp)
 static int
 ex_discard(SCR *sp)
 {
-	WIN *wp;
+	GS *gp;
 	EXCMD *ecp;
 	RANGE *rp;
 
@@ -2166,18 +2179,26 @@ ex_discard(SCR *sp)
 	 * We know the first command can't be an AGV command, so we don't
 	 * process it specially.  We do, however, nail the command itself.
 	 */
-	for (wp = sp->wp; (ecp = wp->ecq.lh_first) != &wp->excmd;) {
+	for (gp = sp->gp;;) {
+		ecp = SLIST_FIRST(gp->ecq);
+		if (F_ISSET(ecp, E_NAMEDISCARD))
+			free(ecp->if_name);
+		/* Reset the last command without dropping it. */
+		if (ecp == &gp->excmd)
+			break;
 		if (FL_ISSET(ecp->agv_flags, AGV_ALL)) {
-			while ((rp = ecp->rq.cqh_first) != (void *)&ecp->rq) {
-				CIRCLEQ_REMOVE(&ecp->rq, rp, q);
+			while ((rp = TAILQ_FIRST(ecp->rq)) != NULL) {
+				TAILQ_REMOVE(ecp->rq, rp, q);
 				free(rp);
 			}
 			free(ecp->o_cp);
 		}
-		LIST_REMOVE(ecp, q);
+		SLIST_REMOVE_HEAD(gp->ecq, q);
 		free(ecp);
 	}
-	wp->ecq.lh_first->clen = 0;
+
+	ecp->if_name = NULL;
+	ecp->clen = 0;
 	return (0);
 }
 
@@ -2193,7 +2214,7 @@ ex_unknown(SCR *sp, CHAR_T *cmd, size_t len)
 
 	GET_SPACE_GOTOW(sp, bp, blen, len + 1);
 	bp[len] = '\0';
-	MEMCPYW(bp, cmd, len);
+	MEMCPY(bp, cmd, len);
 	msgq_wstr(sp, M_ERR, bp, "098|The %s command is unknown");
 	FREE_SPACEW(sp, bp, blen);
 
@@ -2207,14 +2228,14 @@ alloc_err:
  *	[un]abbreviate command, so it can turn off abbreviations.  See
  *	the usual ranting in the vi/v_txt_ev.c:txt_abbrev() routine.
  *
- * PUBLIC: int ex_is_abbrev __P((SCR *, CHAR_T *, size_t));
+ * PUBLIC: int ex_is_abbrev(CHAR_T *, size_t);
  */
 int
-ex_is_abbrev(SCR *sp, CHAR_T *name, size_t len)
+ex_is_abbrev(CHAR_T *name, size_t len)
 {
 	EXCMDLIST const *cp;
 
-	return ((cp = ex_comm_search(sp, name, len)) != NULL &&
+	return ((cp = ex_comm_search(name, len)) != NULL &&
 	    (cp == &cmds[C_ABBR] || cp == &cmds[C_UNABBREVIATE]));
 }
 
@@ -2224,10 +2245,10 @@ ex_is_abbrev(SCR *sp, CHAR_T *name, size_t len)
  *	unmap command, so it can turn off input mapping.  See the usual
  *	ranting in the vi/v_txt_ev.c:txt_unmap() routine.
  *
- * PUBLIC: int ex_is_unmap __P((SCR *, CHAR_T *, size_t));
+ * PUBLIC: int ex_is_unmap(CHAR_T *, size_t);
  */
 int
-ex_is_unmap(SCR *sp, CHAR_T *name, size_t len)
+ex_is_unmap(CHAR_T *name, size_t len)
 {
 	EXCMDLIST const *cp;
 
@@ -2238,7 +2259,7 @@ ex_is_unmap(SCR *sp, CHAR_T *name, size_t len)
 	if (name[len - 1] != '!')
 		return (0);
 	--len;
-	return ((cp = ex_comm_search(sp, name, len)) != NULL &&
+	return ((cp = ex_comm_search(name, len)) != NULL &&
 	    cp == &cmds[C_UNMAP]);
 }
 
@@ -2247,7 +2268,7 @@ ex_is_unmap(SCR *sp, CHAR_T *name, size_t len)
  *	Search for a command name.
  */
 static EXCMDLIST const *
-ex_comm_search(SCR *sp, CHAR_T *name, size_t len)
+ex_comm_search(CHAR_T *name, size_t len)
 {
 	EXCMDLIST const *cp;
 
@@ -2256,7 +2277,8 @@ ex_comm_search(SCR *sp, CHAR_T *name, size_t len)
 			return (NULL);
 		if (cp->name[0] != name[0])
 			continue;
-		if (!MEMCMP(name, cp->name, len))
+		if (STRLEN(cp->name) >= len &&
+			!MEMCMP(name, cp->name, len))
 			return (cp);
 	}
 	return (NULL);
@@ -2267,12 +2289,12 @@ ex_comm_search(SCR *sp, CHAR_T *name, size_t len)
  *	Display a bad address message.
  *
  * PUBLIC: void ex_badaddr
- * PUBLIC:    __P((SCR *, EXCMDLIST const *, enum badaddr, enum nresult));
+ * PUBLIC:   (SCR *, EXCMDLIST const *, enum badaddr, enum nresult);
  */
 void
 ex_badaddr(SCR *sp, const EXCMDLIST *cp, enum badaddr ba, enum nresult nret)
 {
-	db_recno_t lno;
+	recno_t lno;
 
 	switch (nret) {
 	case NUM_OK:
@@ -2307,7 +2329,7 @@ ex_badaddr(SCR *sp, const EXCMDLIST *cp, enum badaddr ba, enum nresult nret)
 		if (lno != 0) {
 			msgq(sp, M_ERR,
 			    "102|Illegal address: only %lu lines in the file",
-			    lno);
+			    (u_long)lno);
 			break;
 		}
 		/* FALLTHROUGH */
@@ -2318,9 +2340,8 @@ ex_badaddr(SCR *sp, const EXCMDLIST *cp, enum badaddr ba, enum nresult nret)
 		abort();
 		/* NOTREACHED */
 	case A_ZERO:
-		msgq(sp, M_ERR,
-		    "104|The %s command doesn't permit an address of 0",
-		    cp->name);
+		msgq_wstr(sp, M_ERR, cp->name,
+		    "104|The %s command doesn't permit an address of 0");
 		break;
 	}
 	return;
@@ -2336,21 +2357,23 @@ ex_comlog(sp, ecp)
 	SCR *sp;
 	EXCMD *ecp;
 {
-	vtrace(sp, "ecmd: %s", ecp->cmd->name);
+	TRACE(sp, "ecmd: "WS, ecp->cmd->name);
 	if (ecp->addrcnt > 0) {
-		vtrace(sp, " a1 %d", ecp->addr1.lno);
+		TRACE(sp, " a1 %d", ecp->addr1.lno);
 		if (ecp->addrcnt > 1)
-			vtrace(sp, " a2: %d", ecp->addr2.lno);
+			TRACE(sp, " a2: %d", ecp->addr2.lno);
 	}
 	if (ecp->lineno)
-		vtrace(sp, " line %d", ecp->lineno);
+		TRACE(sp, " line %d", ecp->lineno);
 	if (ecp->flags)
-		vtrace(sp, " flags 0x%x", ecp->flags);
-	if (F_ISSET(&exc, E_BUFFER))
-		vtrace(sp, " buffer %c", ecp->buffer);
-	if (ecp->argc)
+		TRACE(sp, " flags 0x%x", ecp->flags);
+	if (FL_ISSET(ecp->iflags, E_C_BUFFER))
+		TRACE(sp, " buffer "WC, ecp->buffer);
+	if (ecp->argc) {
+		int cnt;
 		for (cnt = 0; cnt < ecp->argc; ++cnt)
-			vtrace(sp, " arg %d: {%s}", cnt, ecp->argv[cnt]->bp);
-	vtrace(sp, "\n");
+			TRACE(sp, " arg %d: {"WS"}", cnt, ecp->argv[cnt]->bp);
+	}
+	TRACE(sp, "\n");
 }
 #endif
